######################## 文件简介 ########################
# 实现了两个将客流csv文件转化为模拟器可用的客流数据集的工具函数

import os
import numpy as np
import pickle
import glob
from csv import reader
from offline_tools.preprocess_data import *


class MyQueue:
    def __init__(self):
        self.data = []
        self.idx = 0

    def empty(self):
        return self.idx >= len(self.data)

    def get(self):
        ret = self.data[self.idx]
        self.idx += 1
        return ret

    def put(self, d):
        self.data.append(d)

    def qsize(self):
        return len(self.data) - self.idx

######################## 函数简介 ########################
# 读取csv文件，将csv文件中记录的客流转化为一组模拟器可用的固定的客流
# 输入：csv客流文件
# 输出：模拟器可用的客流，是一个MyQueue类型的队列。
def generate_dataset_from_csv_to_pipline(data_file, dt=0.5, data_of_section=''):
    collected_data = MyQueue()
    used_ele = None
    last_ele = -1
    if data_of_section != '':
        # '10:00-11:00'
        s, e = data_of_section.split('-')
        sm, ss = s.split(':')
        em, es = e.split(':')
        section_start = int(sm) * 60 + float(ss)
        section_end = int(em) * 60 + float(es)
    with open(data_file, 'rt', encoding='utf8', errors='ignore') as f:
        r = reader(f)
        row = None
        while True:
            if row != None and row != [] and row[0] == 'PASSENGER LIST':
                break
            row = next(r)
        # #for i in range(274):
        for i in range(10):
            row = next(r)
        awt, att = 0, 0
        while True:
            time, start_level, end_level, m, standard_ele = row[0], int(row[1].replace('Level ', '')), int(row[2].replace('Level ', '')), float(row[3]),  int(row[7]) - 1
            wt, tt = float(row[10]), float(row[11])
            next_person_time = time.split(':')
            minute, second = next_person_time[-2:]
            nt = int(minute) * 60 + float(second)
            row = next(r)

            if data_of_section != '' and nt < section_start:
                continue
            if data_of_section != '' and nt > section_end:
                break
            collected_data.put({'appear_time': nt, 'start_level': start_level, 'end_level': end_level, 'm': m, 'standard_ele': standard_ele})
            awt += wt
            att += tt
            if last_ele != standard_ele:
                # used_ele.put(standard_ele)
                last_ele = standard_ele

            if len(row) < 14 or row[0]=='':
                break

    awt /= collected_data.qsize()
    att /= collected_data.qsize()
    # print(data_file, awt, att)
    return collected_data, used_ele


######################## 函数简介 ########################
# 读取一组csv文件，提取客流的生成概率，利用生成概率，生成一个随机的客流数据。
# 输入：客流生成概率矩阵。
# 输出：模拟器可用的客流，是一个MyQueue类型的队列。
def generate_dataset_from_prob_to_pipline(generate_mask, data_of_section=''):

    if data_of_section != '':
        # '10:00-11:00'
        s, e = data_of_section.split('-')
        sm, ss = s.split(':')
        em, es = e.split(':')
        section_start = int(sm) * 60 + int(ss)
        section_end = int(em) * 60 + int(es)
        t_range = range(section_start, section_end)
    else:
        t_range = range(generate_mask.shape[0])

    collected_data = MyQueue()
    for t in t_range:
        for s in range(generate_mask.shape[1]):
            for d in range(generate_mask.shape[2]):
                if generate_mask[t][s][d]:
                    collected_data.put({'appear_time': t, 'start_level': s+1, 'end_level': d+1, 'm': 75, 'standard_ele': -1})
    return collected_data


def save_dataset(data, file):
    out_put = open(file, 'wb')
    data_str = pickle.dumps(data)
    out_put.write(data_str)
    out_put.close()


def load_dataset(file):
    with open(file, 'rb') as file:
        data = pickle.loads(file.read())
    return data



