import os
from csv import reader, writer
import numpy as np


######################## 函数简介 ########################
# csv文件数据预处理，将csv文件转化为乘客生成概率矩阵。
# 输入：csv客流文件
# 输出：生成概率矩阵
def process2(data_dir='./down_data', floor_num=16, dt=60, t=3600):
    iter_num = t // dt
    data_x = []
    max_value = 0.0
    files = os.listdir(data_dir)
    for file in files:
        data_file = os.path.join(data_dir, file)
        if not data_file.endswith('.csv'):
            continue
        f_t_s_d = [[[0 for k in range(floor_num)] for j in range(floor_num)] for i in range(iter_num)]
        with open(data_file, 'rt', encoding='utf8', errors='ignore') as f:
            r = reader(f)
            row = None
            while True:
                if row != None and row != [] and row[0] == 'PASSENGER LIST':
                    break
                row = next(r)
            for i in range(10):
                row = next(r)
            while True:
                time, start_level, end_level = row[0], int(row[1].replace('Level ', '')) - 1, int(
                    row[2].replace('Level ', '')) - 1
                next_person_time = time.split(':')
                nt = int(next_person_time[-2]) * 60 + float(next_person_time[-1])
                # print(nt, int(nt)//dt)
                f_t_s_d[int(nt) // dt][start_level][end_level] += 1
                row = next(r)
                if len(row) < 14 or row[0] == '':
                    break

        # print(start_data)
        tmp_max = np.max(f_t_s_d)
        if tmp_max > max_value:
            max_value = tmp_max
        data_x.append(f_t_s_d)
    # return np.asarray(data_x) / max_value, np.asarray(data_y) / max_value
    return np.asarray(data_x)


######################## 函数简介 ########################
# csv文件数据预处理，将csv文件转化为乘客生成概率矩阵。
# 与process2的区别主要是，process2默认客流csv文件都是一个小时时长的，而本函数是超过一个小时的
# 输入：csv客流文件
# 输出：生成概率矩阵
def process3(data_dir='./down_data', floor_num=16, dt=60, t=3600):
    iter_num = t // dt
    data_x = []
    max_value = 0.0
    files = os.listdir(data_dir)
    first_hour = -1
    for file in files:
        data_file = os.path.join(data_dir, file)
        if not data_file.endswith('.csv'):
            continue
        f_t_s_d = [[[0 for k in range(floor_num)] for j in range(floor_num)] for i in range(iter_num)]
        with open(data_file, 'rt', encoding='utf8', errors='ignore') as f:
            r = reader(f)
            row = None
            while True:
                if row != None and row != [] and row[0] == 'PASSENGER LIST':
                    break
                row = next(r)
            for i in range(10):
                row = next(r)
            while True:
                time, start_level, end_level = row[0], int(row[1].replace('Level ', '')) - 1, int(
                    row[2].replace('Level ', '')) - 1
                next_person_time = time.split(':')
                nt = int(next_person_time[-2]) * 60 + float(next_person_time[-1])
                if len(next_person_time) == 3:
                    # nt += int(next_person_time[0]) * 3600
                    if first_hour == -1:
                        first_hour = next_person_time[0]
                    elif first_hour != next_person_time[0]:
                        break
                # print(nt, int(nt)//dt)
                f_t_s_d[int(nt) // dt][start_level][end_level] += 1
                row = next(r)
                if len(row) < 14 or row[0] == '':
                    break

        # print(start_data)
        tmp_max = np.max(f_t_s_d)
        if tmp_max > max_value:
            max_value = tmp_max
        data_x.append(f_t_s_d)
    # return np.asarray(data_x) / max_value, np.asarray(data_y) / max_value
    return np.asarray(data_x)